from fastapi import HTTPException
from sumy.summarizers.lex_rank import LexRankSummarizer
from sumy.summarizers.lsa import LsaSummarizer
from sumy.summarizers.luhn import LuhnSummarizer
from sumy.nlp.tokenizers import Tokenizer
from sumy.parsers.plaintext import PlaintextParser
from jaseci.actions.live_actions import jaseci_action
from sumy.parsers.html import HtmlParser
from sumy.nlp.stemmers import Stemmer
from sumy.utils import get_stop_words
import nltk

nltk.download("punkt")
nltk.download("stopwords")

# summarizes the text / url passed to the endpoint and
# returns summary based on the sentences.


@jaseci_action(act_group=["cl_summer"], allow_remote=True)
def summarize(
    text: str = "none",
    url: str = "none",
    sent_count: int = 1,
    summarizer_type: str = "LsaSummarizer",
):
    LANGUAGE = "english"

    # checking which summarization models to use
    if summarizer_type == "LsaSummarizer":
        summarizer_type = LsaSummarizer

    elif summarizer_type == "LexRankSummarizer":
        summarizer_type = LexRankSummarizer

    elif summarizer_type == "LuhnSummarizer":
        summarizer_type = LuhnSummarizer

    else:
        raise HTTPException(
            status_code=404,
            detail=str(
                """Supported type of summarizers are
                LexRankSummarizer / LsaSummarizer / LuhnSummarizer."""
            ),
        )

    # checks if text / url key exists on the api payload
    if text != "none" and url == "none":
        parser = PlaintextParser.from_string(text, Tokenizer(LANGUAGE))
    elif text == "none" and url != "none":
        parser = HtmlParser.from_url(url, Tokenizer(LANGUAGE))
    elif text != "none" and url != "none":
        raise HTTPException(
            status_code=404,
            detail=str("only one of the following key is required text / url."),
        )
    else:
        raise HTTPException(
            status_code=404,
            detail=str("one of the following key is required text / url."),
        )

    sentences = []

    stemmer = Stemmer(LANGUAGE)
    summarizer = summarizer_type(stemmer)
    summarizer.stop_words = get_stop_words(LANGUAGE)

    for sentence in summarizer(parser.document, sent_count):
        sentences.append(str(sentence))

    return sentences


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



from jaseci.utils.test_core import CoreTest, jac_testcase
from jaseci.actions.live_actions import load_module_actions, unload_module
import pytest


class BiNERTest(CoreTest):
    fixture_src = __file__

    @classmethod
    def setUpClass(cls):
        super(BiNERTest, cls).setUpClass()
        ret = load_module_actions("jac_nlp.bi_ner")
        assert ret is True

    @pytest.mark.order(1)
    @jac_testcase("bi_ner.jac", "test_bi_ner_train")
    def test_ner_train(self, ret):
        print(ret)
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(2)
    @jac_testcase("bi_ner.jac", "test_bi_ner_infer")
    def test_biner_infer(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(3)
    @jac_testcase("bi_ner.jac", "test_bi_ner_save_model")
    def test_biencoder_save_model(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(4)
    @jac_testcase("bi_ner.jac", "test_bi_ner_load_model")
    def test_biencoder_load_model(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(5)
    @jac_testcase("bi_ner.jac", "test_bi_ner_set_model_config")
    def test_biencoder_set_model_config(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(6)
    @jac_testcase("bi_ner.jac", "test_bi_ner_set_train_config")
    def test_biencoder_set_train_config(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(7)
    @jac_testcase("bi_ner.jac", "test_bi_ner_get_train_config")
    def test_biencoder_get_train_config(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(8)
    @jac_testcase("bi_ner.jac", "test_bi_ner_get_model_config")
    def test_biencoder_get_model_config(self, ret):
        self.assertEqual(ret["success"], True)

    @classmethod
    def tearDownClass(cls):
        super(BiNERTest, cls).tearDownClass()
        ret = unload_module("jac_nlp.bi_ner.bi_ner")
        assert ret is True



from transformers import (
    AutoTokenizer,
    DataCollatorForTokenClassification,
    AutoModelForTokenClassification,
    TrainingArguments,
    Trainer,
)
from transformers import pipeline
from torch import cuda
from datetime import datetime
import numpy as np
import evaluate
import re

device = "cuda" if cuda.is_available() else "cpu"
print("Using device for training -> ", device)


# Logging
def logs(*args):
    with open("train/logs/" + args[-1], "a") as f:
        data = ""
        for arg in args[:-1]:
            data += str(arg)
        print(data)
        f.write(data)
        f.write("\n")


metric = evaluate.load("seqeval")
model = None


def compute_metrics(eval_preds):
    logits, labels = eval_preds
    predictions = np.argmax(logits, axis=-1)

    # Remove ignored index (special tokens) and convert to labels
    true_labels = [
        [val_dm.unique_entities[lbl] for lbl in label if lbl != -100]
        for label in labels
    ]
    true_predictions = [
        [
            val_dm.unique_entities[p]
            for (p, lbl) in zip(prediction, label)
            if lbl != -100
        ]
        for prediction, label in zip(predictions, labels)
    ]
    all_metrics = metric.compute(predictions=true_predictions, references=true_labels)
    return {
        "precision": all_metrics["overall_precision"],
        "recall": all_metrics["overall_recall"],
        "f1": all_metrics["overall_f1"],
        "accuracy": all_metrics["overall_accuracy"],
    }


def get_tokens_with_entities(raw_text: str):
    # split the text by spaces only if the space does not occur between square brackets
    # we do not want to split "multi-word" entity value yet
    raw_tokens = re.split(r"\s(?![^\[]*\])", raw_text)

    # a regex for matching the annotation  [entity_value](entity_name)
    entity_value_pattern = r"\[(?P<value>.+?)\]\((?P<entity>.+?)\)"
    entity_value_pattern_compiled = re.compile(entity_value_pattern, flags=re.I | re.M)

    tokens_with_entities = []

    for raw_token in raw_tokens:
        match = entity_value_pattern_compiled.match(raw_token)
        if match:
            raw_entity_name, raw_entity_value = match.group("entity"), match.group(
                "value"
            )

            for i, raw_entity_token in enumerate(re.split("\s", raw_entity_value)):
                entity_prefix = "B" if i == 0 else "I"
                entity_name = f"{entity_prefix}-{raw_entity_name}"
                tokens_with_entities.append((raw_entity_token, entity_name))
        else:
            tokens_with_entities.append((raw_token, "O"))

    return tokens_with_entities


class NERDataMaker:
    def __init__(self, texts):
        self.unique_entities = []
        self.processed_texts = []

        temp_processed_texts = []
        for text in texts:
            tokens_with_entities = get_tokens_with_entities(text)
            for _, ent in tokens_with_entities:
                if ent not in self.unique_entities:
                    self.unique_entities.append(ent)
            temp_processed_texts.append(tokens_with_entities)

        self.unique_entities.sort(key=lambda ent: ent if ent != "O" else "")

        for tokens_with_entities in temp_processed_texts:
            self.processed_texts.append(
                [
                    (t, self.unique_entities.index(ent))
                    for t, ent in tokens_with_entities
                ]
            )

    @property
    def id2label(self):
        return dict(enumerate(self.unique_entities))

    @property
    def label2id(self):
        return {v: k for k, v in self.id2label.items()}

    def __len__(self):
        return len(self.processed_texts)

    def __getitem__(self, idx):
        def _process_tokens_for_one_text(id, tokens_with_encoded_entities):
            ner_tags = []
            tokens = []
            for t, ent in tokens_with_encoded_entities:
                ner_tags.append(ent)
                tokens.append(t)

            return {"id": id, "ner_tags": ner_tags, "tokens": tokens}

        tokens_with_encoded_entities = self.processed_texts[idx]
        if isinstance(idx, int):
            return _process_tokens_for_one_text(idx, tokens_with_encoded_entities)
        else:
            return [
                _process_tokens_for_one_text(i + idx.start, tee)
                for i, tee in enumerate(tokens_with_encoded_entities)
            ]

    def as_hf_dataset(self, tokenizer):
        from datasets import Dataset, Features, Value, ClassLabel, Sequence

        def tokenize_and_align_labels(examples):
            tokenized_inputs = tokenizer(
                examples["tokens"], truncation=True, is_split_into_words=True
            )

            labels = []
            for i, label in enumerate(examples["ner_tags"]):
                word_ids = tokenized_inputs.word_ids(
                    batch_index=i
                )  # Map tokens to their respective word.
                previous_word_idx = None
                label_ids = []
                for word_idx in word_ids:  # Set the special tokens to -100.
                    if word_idx is None:
                        label_ids.append(-100)
                    elif (
                        word_idx != previous_word_idx
                    ):  # Only label the first token of a given word.
                        label_ids.append(label[word_idx])
                    else:
                        label_ids.append(-100)
                    previous_word_idx = word_idx
                labels.append(label_ids)

            tokenized_inputs["labels"] = labels
            return tokenized_inputs

        ids, ner_tags, tokens = [], [], []
        for i, pt in enumerate(self.processed_texts):
            ids.append(i)
            pt_tokens, pt_tags = list(zip(*pt))
            ner_tags.append(pt_tags)
            tokens.append(pt_tokens)
        data = {"id": ids, "ner_tags": ner_tags, "tokens": tokens}
        features = Features(
            {
                "tokens": Sequence(Value("string")),
                "ner_tags": Sequence(ClassLabel(names=self.unique_entities)),
                "id": Value("int32"),
            }
        )
        ds = Dataset.from_dict(data, features)
        tokenized_ds = ds.map(tokenize_and_align_labels, batched=True)
        return tokenized_ds


def load_custom_model(model_path, train_dm=None):
    global model, tokenizer, data_collator
    tokenizer = AutoTokenizer.from_pretrained(model_path)
    model = AutoModelForTokenClassification.from_pretrained(
        model_path,
        num_labels=len(train_dm.unique_entities),
        id2label=train_dm.id2label,
        label2id=train_dm.label2id,
    )
    model.to(device)
    data_collator = DataCollatorForTokenClassification(tokenizer=tokenizer)


def train_model(train_data, val_data, train_config):
    global val_dm
    if len(train_data) != 0:
        train_dm = NERDataMaker(train_data)
        train_ds = train_dm.as_hf_dataset(tokenizer=tokenizer)
        print(f"total Train examples = {len(train_data)}")
    if len(val_data) != 0:
        val_dm = NERDataMaker(val_data)
        val_ds = val_dm.as_hf_dataset(tokenizer=tokenizer)
        print(f"total validation examples = {len(val_dm)}")
    else:
        val_dm = train_dm
        val_ds = train_ds
    training_args = TrainingArguments(
        output_dir="./results",
        evaluation_strategy="epoch",
        learning_rate=train_config["LEARNING_RATE"],
        per_device_train_batch_size=train_config["TRAIN_BATCH_SIZE"],
        per_device_eval_batch_size=train_config["VALID_BATCH_SIZE"],
        num_train_epochs=train_config["EPOCHS"],
        weight_decay=0.01,
    )
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_ds,
        eval_dataset=val_ds,
        tokenizer=tokenizer,
        data_collator=data_collator,
        compute_metrics=compute_metrics,
        callbacks=None,
    )

    trainer.train()


def save_custom_model(model_path):
    # saving model
    model.save_pretrained(model_path)
    tokenizer.save_pretrained(model_path)
    print(str(datetime.now()) + "   ", f"model saved successful to : {model_path}")


# predicting entities
def predict_text(sentence):
    if model is not None:
        pipe = pipeline(
            "ner",
            model=model.to("cpu"),
            tokenizer=tokenizer,
            aggregation_strategy="first",
        )
        entities = pipe(sentence)
        ents = []
        for itm in entities:
            ents.append(
                {
                    "entity_text": itm["word"],
                    "entity_value": itm["entity_group"],
                    "conf_score": float(itm["score"]),
                    "start_pos": itm["start"],
                    "end_pos": itm["end"],
                }
            )
        return ents
    return "Model not trained"



from typing import Iterable
from torch.utils.data import Dataset
from .example import Example


class NERDataset(Dataset[Example]):
    def __init__(self, examples: Iterable[Example]):
        self._examples = list(examples)

    def __getitem__(self, index) -> Example:
        return self._examples[index]

    def __len__(self):
        return len(self._examples)



import os

# from pathlib import Path

# import fast_enc


def make_path(file):
    return os.path.join(os.path.dirname(__file__), file)


model_dir = make_path("pretrained")
if not os.path.isdir(model_dir):
    model_dir = make_path("pretrained")
    os.mkdir(model_dir)
model_path = make_path(model_dir + "/model.ftz")
model_file_path = make_path(model_path)
train_file_path = make_path(model_dir + "/train.txt")
test_file_path = make_path("test_sentences.txt")
clf_json_file_path = make_path(model_dir + "/training_data.json")
base_json_file_path = make_path(model_dir + "/base_model.json")



from transformers import pipeline, GPT2Tokenizer, GPT2Model
import torch
from jaseci.actions.live_actions import jaseci_action
import traceback
from fastapi import HTTPException
from typing import List, Union

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

model = None
tokenizer = None
generator = None


def setup(model_name: str = "gpt2-medium", get_embeddings: bool = False):
    global model, tokenizer, generator
    if get_embeddings:
        tokenizer = GPT2Tokenizer.from_pretrained(model_name)
        tokenizer.pad_token = tokenizer.eos_token
        model = GPT2Model.from_pretrained(model_name).to(device)
    else:
        generator = pipeline("text-generation", model=model_name)


setup(model_name="gpt2-medium", get_embeddings=False)


@jaseci_action(act_group=["gpt2"], allow_remote=True)
def generate(
    text: Union[List[str], str],
    max_length: int = 30,
    min_length: int = 10,
    num_return_sequences: int = 3,
) -> List:
    global generator, model, tokenizer
    if generator is None:
        model, tokenizer = None, None
        setup(model_name="gpt2-medium", get_embeddings=False)
    try:
        if isinstance(text, str):
            text = [text]
        return generator(
            text,
            max_length=max_length,
            min_length=min_length,
            num_return_sequences=num_return_sequences,
        )
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["gpt2"], allow_remote=True)
def get_embeddings(text: Union[List[str], str]) -> List:
    global model, tokenizer, generator, device
    if model is None or tokenizer is None:
        generator = None
        setup(model_name="gpt2-medium", get_embeddings=True)
    try:
        if isinstance(text, str):
            text = [text]
        inputs = tokenizer(text, return_tensors="pt", padding=True).to(device)
        with torch.no_grad():
            outputs = model(**inputs)
        return outputs.last_hidden_state.tolist()
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



import spacy
from fastapi import HTTPException
from transformers import AutoTokenizer, AutoModelForSequenceClassification
import torch
from jaseci.actions.live_actions import jaseci_action

# loading segmentation model from hugging face
tokenizer = AutoTokenizer.from_pretrained("dennlinger/roberta-cls-consec")
model = AutoModelForSequenceClassification.from_pretrained(
    "dennlinger/roberta-cls-consec"
)
# Download the pretrained model pipeline
spacy.cli.download("en_core_web_sm")
# loading space model for sentence tokenization
pipeline = spacy.load("en_core_web_sm")


def segmentation(text, threshold=0.85):
    """
    This function takes the raw text, converts them into sentence tokens.
    Tokens are then provided to the sequence classifier to predict the
    syntactical similarity.
    """
    # spliting the raw test into sentences
    doc = pipeline(text)
    sentences = [sent.text.strip() for sent in doc.sents]
    index = 0
    sub_segments = []
    segments = []
    for sent_ind in range(1, len(sentences)):
        # create sentence pair to evaluate
        pair = sentences[index] + " " + sentences[sent_ind]
        inputs = tokenizer.encode_plus(
            pair, max_length=128, pad_to_max_length=True, return_tensors="pt"
        )
        outputs = model(**inputs).logits
        # getting the similarity score between sentences
        value = torch.softmax(outputs, dim=1).tolist()[0][1]
        # creating sub-segment of similar text
        if value > threshold:
            if not sub_segments:
                sub_segments.append(sentences[index])
                sub_segments.append(sentences[sent_ind])
            else:
                sub_segments.append(sentences[sent_ind])
        else:
            if not sub_segments:
                sub_segments.append(sentences[index])
            segments.append(" ".join(sub_segments))
            sub_segments = []
            sub_segments.append(sentences[sent_ind])
        index += 1
    if sub_segments:
        segments.append(" ".join(sub_segments))
    # returning final segments of text
    return segments


@jaseci_action(act_group=["text_seg"], allow_remote=True)
def get_segments(text: str, threshold: float = 0.7):
    try:
        segmented_text = segmentation(text=text, threshold=threshold)
        return segmented_text
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["text_seg"], allow_remote=True)
# NOTE: Temporary fix for jaseci kit CI pipeline.
# Need to change back to load_model for consistency with other models.
def seg_load_model(model_name: str):  # modelname could be ("wiki", "legal")
    global model, tokenizer
    if model_name == "wiki":
        tokenizer = AutoTokenizer.from_pretrained("dennlinger/bert-wiki-paragraphs")
        model = AutoModelForSequenceClassification.from_pretrained(
            "dennlinger/bert-wiki-paragraphs"
        )
    elif model_name == "legal":
        tokenizer = AutoTokenizer.from_pretrained("dennlinger/roberta-cls-consec")
        model = AutoModelForSequenceClassification.from_pretrained(
            "dennlinger/roberta-cls-consec"
        )
    else:
        raise HTTPException(status_code=500, detail="Invalid model name.")
    return f"[Model Loaded] : {model_name}"


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    print("Text Segmentor up and running")
    launch_server(port=8000)



from .action_configs.bart_sum_action_config import BART_SUM_ACTION_CONFIG
from .action_configs.bi_enc_action_config import BI_ENC_ACTION_CONFIG
from .action_configs.bi_ner_action_config import BI_NER_ACTION_CONFIG
from .action_configs.cl_summer_action_config import CL_SUMMER_ACTION_CONFIG
from .action_configs.ent_ext_action_config import ENT_EXT_ACTION_CONFIG
from .action_configs.fast_enc_action_config import FAST_ENC_ACTION_CONFIG
from .action_configs.sbert_sim_action_config import SBERT_SIM_ACTION_CONFIG
from .action_configs.text_seg_action_config import TEXT_SEG_ACTION_CONFIG
from .action_configs.tfm_ner_action_config import TFM_NER_ACTION_CONFIG
from .action_configs.topic_ext_action_config import TOPIC_EXT_ACTION_CONFIG
from .action_configs.use_enc_action_config import USE_ENC_ACTION_CONFIG
from .action_configs.use_qa_action_config import USE_QA_ACTION_CONFIG
from .action_configs.zs_classifier_action_config import ZS_CLASSIFIER_ACTION_CONFIG

ACTION_CONFIGS = {
    "bart_sum": BART_SUM_ACTION_CONFIG,
    "bi_enc": BI_ENC_ACTION_CONFIG,
    "bi_ner": BI_NER_ACTION_CONFIG,
    "cl_summer": CL_SUMMER_ACTION_CONFIG,
    "ent_ext": ENT_EXT_ACTION_CONFIG,
    "fast_enc": FAST_ENC_ACTION_CONFIG,
    "sbert_sim": SBERT_SIM_ACTION_CONFIG,
    "text_seg": TEXT_SEG_ACTION_CONFIG,
    "tfm_ner": TFM_NER_ACTION_CONFIG,
    "topic_ext": TOPIC_EXT_ACTION_CONFIG,
    "use_enc": USE_ENC_ACTION_CONFIG,
    "use_qa": USE_QA_ACTION_CONFIG,
    "zs_classifier": ZS_CLASSIFIER_ACTION_CONFIG,
}



from jaseci.utils.test_core import CoreTest, jac_testcase
from jaseci.actions.live_actions import load_module_actions, unload_module


class UseQAModule(CoreTest):
    fixture_src = __file__

    @classmethod
    def setUpClass(cls):
        super(UseQAModule, cls).setUpClass()
        ret = load_module_actions("jac_nlp.use_qa")
        assert ret == True

    @jac_testcase("use_qa.jac", "test_enc_question_similarity")
    def test_enc_question_similarity(self, ret):
        self.assertGreaterEqual(round(ret["report"][0], 2), 0.9)

    @jac_testcase("use_qa.jac", "test_enc_answer_similarity")
    def test_enc_answer_similarity(self, ret):
        self.assertGreaterEqual(round(ret["report"][0], 2), 0.9)

    @jac_testcase("use_qa.jac", "test_enc_question_classify")
    def test_enc_question_classify(self, ret):
        self.assertEqual(ret["report"][0]["match"], "getdirections")

    @jac_testcase("use_qa.jac", "test_enc_answer_classify")
    def test_enc_answer_classify(self, ret):
        self.assertEqual(ret["report"][0]["match"], "searchplace")

    @jac_testcase("use_qa.jac", "test_enc_question")
    def test_enc_question(self, ret):
        self.assertEqual(len(ret["report"][0][0]), 512)

    @jac_testcase("use_qa.jac", "test_enc_answer")
    def test_enc_answer(self, ret):
        self.assertEqual(len(ret["report"][0][0]), 512)

    @jac_testcase("use_qa.jac", "test_enc_qa_classify")
    def test_enc_qa_classify(self, ret):
        self.assertEqual(ret["report"][0]["match"], "searchplace")

    @jac_testcase("use_qa.jac", "test_enc_qa_similarity")
    def test_enc_qa_similarity(self, ret):
        self.assertGreaterEqual(round(ret["report"][0], 2), 0.4)

    @classmethod
    def tearDownClass(cls):
        super(UseQAModule, cls).tearDownClass()
        ret = unload_module("jac_nlp.use_qa.use_qa")
        assert ret == True



import pandas as pd


def data_token(filename):
    f = open(filename, "r", encoding="utf8")
    split_labeled_text = []
    sentence = []
    sent = 0
    for line in f:
        if len(line) == 0 or line.startswith("-DOCSTART") or line[0] == "\n":
            if len(sentence) > 0:
                split_labeled_text.append(sentence)
                sentence = []
                sent += 1
            continue
        splits = line.split(" ")
        sentence.append(("Sentence:" + str(sent), splits[0], splits[-1].rstrip("\n")))
    f.close()
    if len(sentence) > 0:
        split_labeled_text.append(sentence)
        sentence = []

    data = []
    for i in range(len(split_labeled_text)):
        for s in split_labeled_text[i]:
            data.append(s)
    df = pd.DataFrame(data, columns=["Sentence #", "Word", "Tag"])
    return df


def load_data(filename):
    # reading data from file
    data = data_token(filename)

    data["sentence"] = (
        data[["Sentence #", "Word", "Tag"]]
        .groupby(["Sentence #"])["Word"]
        .transform(lambda x: " ".join(x))
    )
    # let's also create a new column called "word_labels"
    # which groups the tags by sentence
    data["word_labels"] = (
        data[["Sentence #", "Word", "Tag"]]
        .groupby(["Sentence #"])["Tag"]
        .transform(lambda x: ",".join(x))
    )

    labels_name = sorted(list(data.Tag.unique()), reverse=True)

    data = data[["sentence", "word_labels"]].drop_duplicates().reset_index(drop=True)
    return data, labels_name



import torch
import torch.nn as nn
from transformers import PreTrainedModel
import torch.nn.functional as F


class PolyEncoder(PreTrainedModel):
    def __init__(self, config, *inputs, **kwargs):
        super().__init__(config, *inputs, **kwargs)
        if kwargs["shared"] is True:
            self.cont_bert = kwargs["cont_bert"]
            self.cand_bert = self.cont_bert
        else:
            self.cont_bert = kwargs["cont_bert"]
            self.cand_bert = kwargs["cand_bert"]
        self.input_size = None
        self.l_type = kwargs["loss_type"]
        self.l_func = kwargs["loss_function"]
        self.poly_m = kwargs["poly_m"]
        self.poly_code_embeddings = nn.Embedding(self.poly_m, config.hidden_size)
        # https://github.com/facebookresearch/ParlAI/blob/master/parlai/agents/transformer/polyencoder.py#L355 # noqa
        torch.nn.init.normal_(
            self.poly_code_embeddings.weight, config.hidden_size**-0.5
        )

    def dot_attention(self, q, k, v):
        # q: [bs, poly_m, dim] or [bs, res_cnt, dim]
        # k=v: [bs, length, dim] or [bs, poly_m, dim]
        attn_weights = torch.matmul(q, k.transpose(2, 1))  # [bs, poly_m, length]
        attn_weights = F.softmax(attn_weights, -1)
        output = torch.matmul(attn_weights, v)  # [bs, poly_m, dim]
        return output

    def forward(
        self,
        context_input_ids=None,
        context_input_masks=None,
        candidate_input_ids=None,
        candidate_input_masks=None,
        labels=None,
        get_embedding=None,
        mode="train",
        pooling="mean",
    ):
        if get_embedding == "context" or mode == "train" or mode == "eval":
            # batch_size, res_cnt, seq_length = context_input_ids.shape
            # context encoder
            context_vec = self.cont_bert(context_input_ids, context_input_masks)[0]
            if pooling == "mean":
                poly_code_ids = torch.arange(self.poly_m, dtype=torch.long).to(
                    context_input_ids.device
                )
                poly_code_ids = poly_code_ids.unsqueeze(0).expand(
                    context_input_ids.shape[0], self.poly_m
                )
                poly_codes = self.poly_code_embeddings(
                    poly_code_ids
                )  # [bs, poly_m, dim]
                context_vec = self.dot_attention(
                    poly_codes, context_vec, context_vec
                )  # [bs, poly_m, dim]
                if get_embedding == "context":
                    return context_vec
        if get_embedding == "candidate" or mode == "train" or mode == "eval":
            batch_size, res_cnt, seq_length = candidate_input_masks.shape
            candidate_input_ids = candidate_input_ids.view(-1, seq_length)
            candidate_input_masks = candidate_input_masks.view(-1, seq_length)
            candidate_vec = self.cand_bert(candidate_input_ids, candidate_input_masks)[
                0
            ]
            if pooling == "mean":
                # Mean pooling
                output_vectors = []
                input_mask_expanded = (
                    candidate_input_masks.unsqueeze(-1)
                    .expand(candidate_vec.size())
                    .float()
                )
                sum_embeddings = torch.sum(candidate_vec * input_mask_expanded, 1)
                sum_mask = input_mask_expanded.sum(1)
                sum_mask = torch.clamp(sum_mask, min=1e-9)
                output_vectors.append(sum_embeddings / sum_mask)
                candidate_vec = torch.cat(output_vectors, 1)
                candidate_vec = candidate_vec.view(batch_size, res_cnt, -1)
                if get_embedding == "candidate":
                    return candidate_vec
        # merge
        if labels is not None:
            # we are recycling candidates for faster training
            # we repeat candidates for batch_size times to simulate test phase
            # so that every context is paired with batch_size candidates
            candidate_vec = candidate_vec.permute(1, 0, 2)  # [1, bs, dim]
            candidate_vec = candidate_vec.expand(
                batch_size, batch_size, candidate_vec.shape[2]
            )
            context_vec = self.dot_attention(
                candidate_vec, context_vec, context_vec
            ).squeeze()
            dot_product = (context_vec * candidate_vec).sum(-1)  # [bs, bs]
            mask = torch.eye(batch_size).to(context_input_ids.device)  # [bs, bs]
            dot_product = 1 - F.log_softmax(dot_product, dim=-1) * mask
            # loss = (-loss.sum(dim=1)).mean()
            loss_fnct = nn.MSELoss()
            loss = loss_fnct(dot_product.mean(dim=-1), target=labels.squeeze(1))
            return loss
        else:
            context_vec = self.dot_attention(
                candidate_vec, context_vec, context_vec
            )  # [bs, res_cnt, dim]
            dot_product = (context_vec * candidate_vec).sum(-1)
            return dot_product


class BiEncoder(PreTrainedModel):
    def __init__(self, config, *inputs, **kwargs):
        super().__init__(config, *inputs, **kwargs)
        # if shared is true it creates only one model (Siamese type)
        if kwargs["shared"] is True:
            self.cont_bert = kwargs["cont_bert"]
            self.cand_bert = self.cont_bert
        else:
            self.cont_bert = kwargs["cont_bert"]
            self.cand_bert = kwargs["cand_bert"]
        self.input_size = None
        self.l_type = kwargs["loss_type"]
        self.l_func = kwargs["loss_function"]

    def loss_func(self, candidate_vec, context_vec, labels, l_type="cos", l_func="MSE"):
        candidate_vec = candidate_vec.squeeze(1)
        labels = labels.squeeze(1)
        if l_type == "dot":
            dot_product = torch.matmul(context_vec, candidate_vec.t())
            mask = torch.eye(self.input_size).to(candidate_vec.device)
            dot_product = 1 - F.log_softmax(dot_product, dim=-1) * mask
            if l_func == "contrastive":
                dot_product = dot_product.mean(dim=-1)
                loss = 0.5 * (
                    labels.float() * dot_product.pow(2)
                    + (1 - labels).float() * F.relu(0.5 - dot_product).pow(2)
                )
                loss = loss.mean()
            else:
                loss_fnct = nn.MSELoss()
                loss = loss_fnct(dot_product.mean(dim=-1), target=labels)
        else:
            cos = nn.CosineSimilarity()
            if l_func == "contrastive":
                loss_angle = 1 - cos(context_vec, candidate_vec)
                loss = 0.5 * (
                    labels.float() * loss_angle.pow(2)
                    + (1 - labels).float() * F.relu(0.5 - loss_angle).pow(2)
                )
                loss = loss.mean(dim=-1)
            else:
                loss_angle = cos(context_vec, candidate_vec)
                loss_fnct = nn.MSELoss()
                loss = loss_fnct(loss_angle, target=labels)
        return loss

    def forward(
        self,
        context_input_ids=None,
        context_input_masks=None,
        candidate_input_ids=None,
        candidate_input_masks=None,
        labels=None,
        get_embedding=None,
        mode="train",
        pooling="mean",
    ):
        # only select the first candidate (whose lbl==1)
        # if labels is not None:
        #   candidate_input_ids = candidate_input_ids[:, 0, :].unsqueeze(1)
        #   candidate_input_masks = candidate_input_masks[:, 0, :].unsqueeze(1)
        # gets the context embedding
        if get_embedding == "context" or mode == "train" or mode == "eval":
            self.input_size = context_input_ids.size(0)
            context_vec = self.cont_bert(context_input_ids, context_input_masks)[0]
            if pooling == "mean":
                # Mean pooling
                output_vectors = []
                input_mask_expanded = (
                    context_input_masks.unsqueeze(-1).expand(context_vec.size()).float()
                )
                sum_embeddings = torch.sum(context_vec * input_mask_expanded, 1)
                sum_mask = input_mask_expanded.sum(1)
                sum_mask = torch.clamp(sum_mask, min=1e-9)
                output_vectors.append(sum_embeddings / sum_mask)
                context_vec = torch.cat(output_vectors, 1)
                if get_embedding == "context":
                    return context_vec
        # gets the candidate embedding
        if get_embedding == "candidate" or mode == "train" or mode == "eval":
            batch_size, res_cnt, seq_length = candidate_input_ids.shape
            candidate_input_ids = candidate_input_ids.view(-1, seq_length)
            candidate_input_masks = candidate_input_masks.view(-1, seq_length)
            candidate_vec = self.cand_bert(candidate_input_ids, candidate_input_masks)[
                0
            ]
            if pooling == "mean":
                # Mean pooling
                output_vectors = []
                input_mask_expanded = (
                    candidate_input_masks.unsqueeze(-1)
                    .expand(candidate_vec.size())
                    .float()
                )
                sum_embeddings = torch.sum(candidate_vec * input_mask_expanded, 1)
                sum_mask = input_mask_expanded.sum(1)
                sum_mask = torch.clamp(sum_mask, min=1e-9)
                output_vectors.append(sum_embeddings / sum_mask)
                candidate_vec = torch.cat(output_vectors, 1)
                candidate_vec = candidate_vec.view(batch_size, res_cnt, -1)
                if get_embedding == "candidate":
                    return candidate_vec
        if labels is not None and mode == "train":
            return self.loss_func(
                candidate_vec=candidate_vec,
                context_vec=context_vec,
                labels=labels,
                l_type=self.l_type,
                l_func=self.l_func,
            )



from jaseci.utils.test_core import CoreTest, jac_testcase
from jaseci.actions.live_actions import load_module_actions, unload_module


class UseEncModule(CoreTest):
    fixture_src = __file__

    @classmethod
    def setUpClass(cls):
        super(UseEncModule, cls).setUpClass()
        ret = load_module_actions("jac_nlp.use_enc")
        assert ret == True

    @jac_testcase("use_enc.jac", "test_enc_cos_sim_score")
    def test_enc_cos_sim_score(self, ret):
        self.assertGreaterEqual(round(ret["report"][0], 2), 0.9)

    @jac_testcase("use_enc.jac", "test_enc_text_similarity")
    def test_enc_text_similarity(self, ret):
        self.assertEqual(round(ret["report"][0], 2), 0.03)

    @jac_testcase("use_enc.jac", "test_enc_text_classify")
    def test_enc_text_classify(self, ret):
        self.assertEqual(ret["report"][0]["match"], "getdirections")

    @jac_testcase("use_enc.jac", "test_enc_get_embeddings")
    def test_enc_get_embeddings(self, ret):
        self.assertEqual(len(ret["report"][0][0]), 512)

    @classmethod
    def tearDownClass(cls):
        super(UseEncModule, cls).tearDownClass()
        ret = unload_module("jac_nlp.use_enc.use_enc")
        assert ret == True



from typing import List, Union

import numpy as np
from fastapi import HTTPException
from flair.data import Sentence
from flair.models import TARSClassifier
from jaseci.actions.live_actions import jaseci_action


def init_model():
    """load the tars classifier for ZS classification"""
    global classifier
    model_name = "tars-base"
    classifier = TARSClassifier.load(model_name)
    print(f"loaded mode : [{model_name}]")


# initialize the classifier
init_model()


# API for getting the cosine similarity
@jaseci_action(act_group=["bi_enc"], allow_remote=True)
def cosine_sim(vec_a: List[float], vec_b: List[float]):
    """
    Caculate the cosine similarity score of two given vectors
    Param 1 - First vector
    Param 2 - Second vector
    Return - float between 0 and 1
    """

    result = np.dot(vec_a, vec_b) / (np.linalg.norm(vec_a) * np.linalg.norm(vec_b))
    return result.astype(float)


# defining the api for ZS classification
@jaseci_action(act_group=["zs_classifier"], allow_remote=True)
def classify(text: Union[str, List[str]], classes: List[str]):
    """
    API for classifying text among classes provided
    """
    response_data_format = []
    try:
        if isinstance(text, list):
            for sent_text in text:
                resp_data = {}
                sent = Sentence(sent_text)
                # predicting class for the text
                classifier.predict_zero_shot(sent, classes)
                pred_output = sent.to_dict()
                resp_data[pred_output["text"]] = pred_output["all labels"]
                response_data_format.append(resp_data)
        else:
            sent = Sentence(text)
            classifier.predict_zero_shot(sent, classes)
            pred_output = sent.to_dict()
            resp_data = {}
            resp_data[pred_output["text"]] = pred_output["all labels"]
            response_data_format.append(resp_data)
        return response_data_format
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(f"Exception :{e}"))


@jaseci_action(act_group=["zs_classifier"], allow_remote=True)
def get_embeddings(texts: Union[str, List[str]]):
    """
    API to get embbeddings for text
    """
    embedder = classifier.tars_embeddings
    if isinstance(texts, str):
        label_sentences = Sentence([texts])
        embedder.embed(label_sentences)
        return label_sentences.get_embedding().cpu().detach().numpy().tolist()
    else:
        embedding_list = []
        for text in texts:
            label_sentences = Sentence(text)
            embedder.embed(label_sentences)
            embedding_list.append(
                label_sentences.get_embedding().cpu().detach().numpy().tolist()
            )
        return embedding_list


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



from typing import Dict, List
from sentence_transformers import InputExample, losses
from sentence_transformers import SentenceTransformer, models
from sentence_transformers.util import cos_sim, dot_score, semantic_search
from torch import nn
from torch.utils.data import DataLoader
import torch
import math
from datetime import datetime
import numpy as np
from fastapi import HTTPException
from jaseci.actions.live_actions import jaseci_action

"""
Declaring the training config
"""
training_config: Dict = {
    "device": "cuda" if torch.cuda.is_available() else "cpu",
    "num_epochs": 2,
    "model_save_path": (
        "output/sent_model" + "-" + datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    ),
    "model_name": "bert-base-uncased",
}


def create_model(model_name="bert-base-uncased", max_seq_length=256):
    """
    This fumctions can be used to create a custom sbert model from transformers
    model name :- valid transformer model from the hugging face hub
    max_seq_len :- the max_seq_len that it would support for tokenization
    Returns :- returns the custom tranformer model in sbert setup
    """
    word_embedding_model = models.Transformer(model_name, max_seq_length=max_seq_length)
    pooling_model = models.Pooling(word_embedding_model.get_word_embedding_dimension())
    dense_model = models.Dense(
        in_features=pooling_model.get_sentence_embedding_dimension(),
        out_features=max_seq_length,
        activation_function=nn.Tanh(),
    )

    model = SentenceTransformer(
        modules=[word_embedding_model, pooling_model, dense_model]
    )
    return model


@jaseci_action(act_group=["sbert_sim"], allow_remote=True)
def train(dataset: List, training_parameters: Dict = None):
    """
    this train function can be used to train the Sbert model to improve
    the sentence similarity.
    train data :- takes list of sentence1 and sentence2 followed by the similarity score
    training_parameter :- take the dictionary of parameters that can be used to
    manipulate the training
    Returns :- confirmations when training is completed
    """
    global model, training_config
    try:
        training_config.update(training_parameters)
        gold_samples = []
        for t_data in dataset:
            inp_example = InputExample(texts=[t_data[0], t_data[1]], label=t_data[2])
            gold_samples.append(inp_example)
        # Define your train dataset, the dataloader and the train loss
        train_dataloader = DataLoader(gold_samples, shuffle=True, batch_size=16)
        train_loss = losses.CosineSimilarityLoss(model)
        # Configure the training.
        # 10% of train data for warm-up
        warmup_steps = math.ceil(
            len(train_dataloader) * training_config["num_epochs"] * 0.1
        )
        print("Warmup-steps: {}".format(warmup_steps))
        # Tune the model
        model.fit(
            train_objectives=[(train_dataloader, train_loss)],
            epochs=training_config["num_epochs"],
            warmup_steps=warmup_steps,
            output_path=training_config["model_save_path"],
        )
        return "Model Training is completed"
    except Exception as e:
        print(e)
        return f"Error Occured {str(e)}"


@jaseci_action(act_group=["sbert_sim"], allow_remote=True)
def get_text_sim(query: List[str], corpus: List[str], top_k=1):
    """
    Calculates the similarity score between query and corpus
    query :- contains a list of queries that is scored against the list of corpus
    corpus :- contain a list of entries that is evaluated with each query
    top_k :- returns k number of top similar entries in corpus
    Returns :- a dictionary with query as key and list of top_k sentences
    from corpus as value
    """
    try:
        resp_sim_matrix = {}
        text_embeddings = model.encode(query)
        label_embeddings = model.encode(corpus)
        all_sim_data = semantic_search(text_embeddings, label_embeddings)
        count = 0
        for sim_data in all_sim_data:
            resp_sim_matrix[query[count]] = []
            for sim_obj in sim_data[:top_k]:
                resp_sim_matrix[query[count]].append(
                    {
                        "text": corpus[sim_obj["corpus_id"]],
                        "sim_score": sim_obj["score"],
                    }
                )
            count += 1
        return resp_sim_matrix
    except Exception as e:
        print(e)
        return f"Error Occured : {str(e)}"


@jaseci_action(act_group=["sbert_sim"], allow_remote=True)
def getembeddings(text: List[str]):
    """
    encodes the list of sentences provided and returns the respected embeddings
    text : contains a list of sentences to generate encodings
    Returns:  List of encoding for each text
    """
    global model
    model.eval()
    try:
        embeddings = model.encode(text)
        return {np.squeeze(np.asarray(embeddings)).tolist()}
    except Exception as e:
        print(e)
        return f"Error Occured : {str(e)}"


@jaseci_action(act_group=["sbert_sim"], allow_remote=True)
def get_cos_score(vec_a: list, vec_b: list):
    """
    Caculate the cosine similarity score of two given vectors
    Param 1 - First vector
    Param 2 - Second vector
    Return - float between 0 and 1
    """
    return np.asarray(cos_sim(vec_a, vec_b)).tolist()


@jaseci_action(act_group=["sbert_sim"], allow_remote=True)
def get_dot_score(vec_a: list, vec_b: list):
    """
    Caculate the dot product of two given vectors
    Param 1 - First vector
    Param 2 - Second vector
    Return - float between 0 and 1
    """
    return np.asarray(dot_score(vec_a, vec_b)).tolist()


@jaseci_action(act_group=["sbert_sim"], allow_remote=True)
def load_model(model_name="all-MiniLM-L12-v2", model_type="default"):
    """
    Load models load in the memory for similarty scoring
    model_type : can be default or tfm_model
            default : loads model from the sbert model zoo
            tfm_model : load tranformer model from the huggingface hub
    model_name : this is name of the model to be loaded
    Returns: confirmation if the model is loaded
    """
    global model
    if model_type == "tfm_model":
        model = create_model(model_name=model_name, max_seq_length=512)
        return f"{model_name} - loaded from huggingface hub"
    elif model_type == "default":
        model = SentenceTransformer(model_name)
        return f"{model_name} - loaded from SBERT library"
    else:
        raise HTTPException(
            status_code=500,
            detail=str(
                """Supported model_type :-  1) 'tfm_model' for transformer based models,
                                            2) 'default' for Sbert models"""
            ),
        )


@jaseci_action(act_group=["sbert_sim"], allow_remote=True)
def get_train_config():
    """
    returns the active values of the training config
    Returns: a dictionary with key as param name and value as current status
    """
    try:
        return training_config
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    load_model()
    launch_server(port=8000)



from jaseci.utils.test_core import CoreTest, jac_testcase
from jaseci.actions.live_actions import load_module_actions, unload_module
import pytest


class SbertSimTest(CoreTest):
    fixture_src = __file__

    @classmethod
    def setUpClass(cls):
        super(SbertSimTest, cls).setUpClass()
        ret = load_module_actions("jac_nlp.sbert_sim")
        assert ret is True

    @pytest.mark.order(1)
    @jac_testcase("sbert_sim.jac", "test_sbert_sim_load_model")
    def test_sbert_sim_load_model(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(2)
    @jac_testcase("sbert_sim.jac", "test_train_sbert_sim")
    def test_train_sbert_sim(self, ret):
        self.assertEqual(ret["success"], True)
        self.assertEqual(ret["report"][0], "Model Training is completed")

    @pytest.mark.order(4)
    @jac_testcase("sbert_sim.jac", "test_get_text_sim")
    def test_get_text_sim(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(5)
    @jac_testcase("sbert_sim.jac", "test_sbert_sim_get_train_config")
    def test_sbert_sim_get_train_config(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(5)
    @jac_testcase("sbert_sim.jac", "test_sbert_sim_get_cos_score")
    def test_sbert_sim_cos_score(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(5)
    @jac_testcase("sbert_sim.jac", "test_sbert_sim_get_dot_score")
    def test_sbert_sim_get_dot_score(self, ret):
        self.assertEqual(ret["success"], True)

    @pytest.mark.order(5)
    @jac_testcase("sbert_sim.jac", "test_sbert_sim_get_embeddings")
    def test_sbert_sim_get_embeddings(self, ret):
        self.assertEqual(ret["success"], True)

    @classmethod
    def tearDownClass(cls):
        super(SbertSimTest, cls).tearDownClass()
        ret = unload_module("jac_nlp.sbert_sim.sbert_sim")
        assert ret is True



from typing import List
from sentence_transformers import InputExample, losses
from sentence_transformers import SentenceTransformer, models
from sentence_transformers.util import cos_sim
from torch import nn
from torch.utils.data import DataLoader
import nlpaug.augmenter.word as naw
import torch
import math
from datetime import datetime
import numpy as np
from fastapi.responses import JSONResponse
from jaseci.actions.live_actions import jaseci_action

model_name = "bert-base-uncased"
device = "cuda" if torch.cuda.is_available() else "cpu"
num_epochs = 2
model_save_path = (
    "output/sent_"
    + model_name.replace("/", "-")
    + "-"
    + datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
)


def create_model(model_name="bert-base-uncased", max_seq_length=256):
    word_embedding_model = models.Transformer(model_name, max_seq_length=max_seq_length)
    pooling_model = models.Pooling(word_embedding_model.get_word_embedding_dimension())
    dense_model = models.Dense(
        in_features=pooling_model.get_sentence_embedding_dimension(),
        out_features=256,
        activation_function=nn.Tanh(),
    )

    model = SentenceTransformer(
        modules=[word_embedding_model, pooling_model, dense_model]
    )
    return model


model = create_model(model_name=model_name, max_seq_length=256)


def get_aug_sample(text1, text2):
    # Synonym replacement using BERT ####
    aug = naw.ContextualWordEmbsAug(
        model_path=model_name, action="insert", device=device
    )
    aug_samples = []
    for sample1, sample2 in zip(text1, text2):
        augmented_texts = aug.augment([sample1, sample2])
        print(augmented_texts)
        inp_example = InputExample(texts=augmented_texts, label=0)
        aug_samples.append(inp_example)
    print("Textual augmentation completed....")
    print("Number of silver pairs generated: {}".format(len(aug_samples)))
    return aug_samples


@jaseci_action(act_group=["sent_enc"], aliases=["train_model"], allow_remote=True)
def train(text1: List[str], text2: List[str]):
    global model, model_name, device

    try:
        gold_samples = []
        for sample1, sample2 in zip(text1, text2):
            inp_example = InputExample(texts=[sample1, sample2], label=1)
            gold_samples.append(inp_example)
        # generate Samples for 0 labels
        aug_samples = get_aug_sample(text1, text2)
        # Define your train dataset, the dataloader and the train loss
        train_dataloader = DataLoader(
            gold_samples + aug_samples, shuffle=True, batch_size=16
        )
        train_loss = losses.ContrastiveLoss(model)
        # Configure the training.
        # 10% of train data for warm-up
        warmup_steps = math.ceil(len(train_dataloader) * num_epochs * 0.1)
        print("Warmup-steps: {}".format(warmup_steps))
        # Tune the model
        model.fit(
            train_objectives=[(train_dataloader, train_loss)],
            epochs=num_epochs,
            warmup_steps=warmup_steps,
            output_path=model_save_path,
        )
        return JSONResponse(content="Model Training is comnpleted", status_code=200)
    except Exception as e:
        print(e)
        return JSONResponse(content=f"Error Occured {str(e)}", status_code=500)


@jaseci_action(act_group=["sent_enc"], aliases=["inference"])
def predict(text1: str, text2: List[str]):
    try:
        sim = np.zeros(len(text2))
        text_embeddings = model.encode(text1)
        label_embeddings = model.encode(text2)
        for i in range(len(text2)):
            sim[i] = cos_sim(text_embeddings, label_embeddings[i])
        print(sim)
        return JSONResponse(
            {"label": text2[np.argmax(sim)], "conf_score": sim[np.argmax(sim)]}
        )
    except Exception as e:
        print(e)
        return JSONResponse(content=f"Error Occured : {str(e)}", status_code=500)


@jaseci_action(act_group=["sent_enc"], aliases=["getembeddings"])
def getEmbedding(text):
    global model
    model.eval()
    try:
        embeddings = model.encode(text)
        return JSONResponse(
            content={"embed": np.squeeze(np.asarray(embeddings)).tolist()},
            status_code=200,
        )
    except Exception as e:
        print(e)
        return JSONResponse(content=f"Error Occured : {str(e)}", status_code=500)


@jaseci_action(act_group=["sent_enc"], aliases=["get_cos_sim"])
def cosine_sim(vec_a: list, vec_b: list, meta):
    """
    Caculate the cosine similarity score of two given vectors
    Param 1 - First vector
    Param 2 - Second vector
    Return - float between 0 and 1
    """
    result = np.dot(vec_a, vec_b) / (np.linalg.norm(vec_a) * np.linalg.norm(vec_b))
    return result.astype(float)



import torch
from jaseci.actions.live_actions import jaseci_action
from transformers import T5Tokenizer, T5ForConditionalGeneration  # , T5Config

# from fastapi import HTTPException

model = T5ForConditionalGeneration.from_pretrained("t5-small")
tokenizer = T5Tokenizer.from_pretrained("t5-small")
device = torch.device("cpu")


# generates summary based on text
def t5_generate_sum(text, min_length, max_length):
    preprocess_text = text.strip().replace("\n", "")
    t5_prepared_Text = "summarize: " + preprocess_text

    tokenized_text = tokenizer.encode(t5_prepared_Text, return_tensors="pt").to(device)

    summary_ids = model.generate(
        tokenized_text,
        num_beams=4,
        no_repeat_ngram_size=2,
        min_length=min_length,
        max_length=max_length,
        early_stopping=True,
    )

    output = tokenizer.decode(summary_ids[0], skip_special_tokens=True)

    return output


# summarize a large body of text using t5 model (small model)
# which returns data at a fast rate.
@jaseci_action(act_group=["t5_sum"], allow_remote=True)
def classify_text(text: str, min_length: int = 30, max_length: int = 100):
    output = t5_generate_sum(text, min_length, max_length)
    return output


if __name__ == "__main__":
    from jaseci.actions.remote_actions import launch_server

    launch_server(port=8000)



from difflib import SequenceMatcher
import os


def matcher(string, pattern):
    """
    Return the start and end index of any pattern present in the text.
    """
    match_list = []
    pattern = pattern.strip()
    seqMatch = SequenceMatcher(None, string, pattern, autojunk=False)
    match = seqMatch.find_longest_match(0, len(string), 0, len(pattern))
    if match.size == len(pattern):
        start = match.a
        end = match.a + match.size
        match_tup = (start, end)
        string = string.replace(pattern, "X" * len(pattern), 1)
        match_list.append(match_tup)

    return match_list, string


def mark_sentence(s, match_list):
    """
    Marks all the entities in the sentence as per the BIO scheme.
    """
    word_dict = {}
    for word in s.split():
        word_dict[word] = "O"

    for start, end, e_type in match_list:
        temp_str = s[start:end]
        tmp_list = temp_str.split()
        if len(tmp_list) > 1:
            word_dict[tmp_list[0]] = "B-" + e_type
            for w in tmp_list[1:]:
                word_dict[w] = "I-" + e_type
        else:
            word_dict[temp_str] = "B-" + e_type
    return word_dict


def clean(text):
    """
    Just a helper fuction to add a space before
    the punctuations for better tokenization
    """
    filters = [
        "!",
        "#",
        "$",
        "%",
        "&",
        "(",
        ")",
        "/",
        "*",
        ".",
        ":",
        ";",
        "<",
        "=",
        ">",
        "?",
        "@",
        "[",
        "\\",
        "]",
        "_",
        "`",
        "{",
        "}",
        "~",
        "'",
        ",",
    ]
    for i in text:
        if i in filters:
            text = text.replace(i, " " + i)

    return text


def create_data(df, filename):
    """
    The function responsible for the creation of data in the said format.
    """
    data_filepath = f"train/{filename}.txt"
    backup_filepath = f"train/{filename}_backup_file.txt"

    if not os.path.exists("train"):
        os.makedirs("train")
    with open(data_filepath, "w", encoding="utf-8") as d_file_p, open(
        backup_filepath, "a"
    ) as b_file_p:
        for text, annotation in zip(df.text, df.annotation):
            match_list = []
            for i in annotation:
                a, text_ = matcher(text, i[0])
                match_list.append((a[0][0], a[0][1], i[1]))

            d = mark_sentence(text, match_list)
            for i in d.keys():
                d_file_p.writelines(i + " " + d[i] + "\n")
                b_file_p.writelines(i + " " + d[i] + "\n")
            d_file_p.writelines("\n")
            b_file_p.writelines("\n")
    return True


def create_data_new(df, filename):
    data_filepath = f"train/{filename}.txt"
    backup_filepath = f"train/{filename}_backup_file.txt"
    if not os.path.exists("train"):
        os.makedirs("train")
    with open(data_filepath, "w", encoding="utf-8") as d_file_p, open(
        backup_filepath, "a"
    ) as b_file_p:
        for text, annotation in zip(df.text, df.annotation):
            text = clean(text)
            text = clean(text)
            split_sent = text.split()
            tags = ["O"] * len(split_sent)
            for i in annotation:
                e_type = i[1]
                ent_val = text[i[2] : i[3]]
                tags[split_sent.index(ent_val.split()[0])] = "B-" + e_type
                split_sent[split_sent.index(ent_val.split()[0])] = (
                    split_sent[split_sent.index(ent_val.split()[0])] + "t"
                )
                if len(ent_val.split()) > 1:
                    for ent in ent_val.split()[1:]:
                        tags[split_sent.index(ent)] = "I-" + e_type
                        split_sent[split_sent.index(ent)] = (
                            split_sent[split_sent.index(ent)] + "t"
                        )
            for w, t in zip(text.split(), tags):
                d_file_p.writelines(w + " " + t + "\n")
                b_file_p.writelines(w + " " + t + "\n")
            d_file_p.writelines("\n")
            b_file_p.writelines("\n")
    return True



from transformers import BartTokenizer, BartForConditionalGeneration
import torch
from typing import List, Union
from jaseci.actions.live_actions import jaseci_action
import traceback
from fastapi import HTTPException
import requests
from bs4 import BeautifulSoup


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

bart_tokenizer = BartTokenizer.from_pretrained("facebook/bart-large-cnn")
bart_model = BartForConditionalGeneration.from_pretrained("facebook/bart-large-cnn").to(
    device
)


@jaseci_action(act_group=["bart_sum"], allow_remote=True)
def summarize(
    text: Union[List[str], str] = None,  # type: ignore
    url: str = None,  # type: ignore
    max_length: int = 100,
    min_length: int = 10,
    num_beams: int = 4,
) -> List[str]:
    try:
        if text is not None:
            if isinstance(text, str):
                text = [text]
        elif url is not None:
            page = requests.get(url)
            soup = BeautifulSoup(page.text, "html.parser")
            text = [" ".join(map(lambda p: p.text, soup.find_all("p")))]
        else:
            raise HTTPException(status_code=400, detail="No text or url provided")

        inputs = bart_tokenizer.batch_encode_plus(
            text, max_length=1024, return_tensors="pt"
        )
        summary_ids = bart_model.generate(
            inputs["input_ids"].to(device),
            num_beams=num_beams,
            max_length=max_length,
            min_length=min_length,
            early_stopping=True,
        )
        return bart_tokenizer.batch_decode(
            summary_ids, skip_special_tokens=True, clean_up_tokenization_spaces=True
        )
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))



import warnings

import pandas as pd

from jaseci.actions.live_actions import jaseci_action
from jaseci.actions.remote_actions import launch_server
from fastapi import HTTPException

from .action_utils import (
    c_tf_idf,
    extract_top_n_words_per_topic,
    generate_topic_label,
    mmr,
)

warnings.filterwarnings("ignore")
warnings.warn("ignore")


@jaseci_action(act_group=["topic_ext"], allow_remote=True)
def topic_extraction(
    texts: list,
    classes: list,
    n_topics: int = 5,
    min_tokens: int = 1,
    max_tokens: int = 2,
):
    if min_tokens > max_tokens:
        raise ValueError(
            "The value of min_tokens shoul be less or equal than the value of max_tokens"
        )
    try:
        doc_label_dict = {"document": texts, "label": classes}
        doc_label_df = pd.DataFrame(doc_label_dict)

        document_table = doc_label_df.groupby(["label"], as_index=False).agg(
            {"document": " ".join}
        )
        tf_idf, count = c_tf_idf(
            document_table.document.values,
            m=len(texts),
            ngram_range=(min_tokens, max_tokens),
        )

        return extract_top_n_words_per_topic(
            tf_idf=tf_idf, count=count, docs_per_topic=document_table, n=n_topics
        )

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["topic_ext"], allow_remote=True)
def headline_generation(texts: list, classes: list):
    try:
        doc_label_dict = {"document": texts, "label": classes}
        doc_label_df = pd.DataFrame(doc_label_dict)

        document_table = (
            doc_label_df.groupby("label")["document"].apply(list).reset_index()
        )
        document_table["headline"] = document_table["document"].apply(
            lambda x: generate_topic_label(x)
        )
        ret_dict = document_table.set_index("label").to_dict()["headline"]
        return ret_dict

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["topic_ext"], allow_remote=True)
def keyword_extraction(
    sentence: str,
    n_words: int,
    min_tokens: int = 1,
    max_tokens: int = 1,
    diversity: float = 0.02,
):
    if min_tokens > max_tokens:
        raise ValueError(
            "The value of min_tokens shoul be less or equal than the value of max_tokens"
        )

    try:
        keywords = mmr(
            input_text=sentence,
            top_n=n_words,
            n_gram_range=(min_tokens, max_tokens),
            diversity=diversity,
        )
        return keywords
    except ValueError as ve:
        print(
            "Try decresing the value of n_words, The n_words should be less or equal to number of unique words in the sentence after removing stopwords"
        )
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    print("The topic extraction module is up and running.")
    launch_server(port=8000)



import torch
from torch.utils.data import DataLoader
from . import tokenizer as token_util


def get_inference(model, tokenizer, contexts, candidates, train_config):
    context_transform = token_util.SelectionJoinTransform(
        tokenizer=tokenizer, max_len=train_config["max_contexts_length"]
    )
    candidate_transform = token_util.SelectionSequentialTransform(
        tokenizer=tokenizer, max_len=train_config["max_candidate_length"]
    )
    labels = [0] * len(candidates)
    test_data = token_util.SelectionDataset(
        contexts=contexts,
        candidates=candidates,
        context_transform=context_transform,
        candidate_transform=candidate_transform,
        mode="eval",
        labels=labels,
    )
    val_dataloader = DataLoader(
        test_data,
        batch_size=train_config["eval_batch_size"],
        collate_fn=test_data.batchify_join_str,
        shuffle=False,
    )
    results = []
    for step, batch in enumerate(val_dataloader, start=1):
        batch = tuple(t.to(train_config["device"]) for t in batch)
        (
            context_token_ids_list_batch,
            context_input_masks_list_batch,
            candidate_token_ids_list_batch,
            candidate_input_masks_list_batch,
            _,
        ) = batch
        with torch.no_grad():
            logits = model(
                context_token_ids_list_batch,
                context_input_masks_list_batch,
                candidate_token_ids_list_batch,
                candidate_input_masks_list_batch,
                mode="eval",
            )
        _, prediction = torch.max(logits, dim=-1)
        prediction = prediction.tolist()
        if isinstance(prediction, list):
            for pred in prediction:
                results.append(candidates[pred])
        else:
            results.append(candidates[prediction])
    return results


# function provides embedding for context and candidate
def get_embeddings(model, tokenizer, text_data, train_config, embed_type="context"):
    if embed_type == "context":
        context_transform = token_util.SelectionJoinTransform(
            tokenizer=tokenizer, max_len=train_config["max_contexts_length"]
        )
        context_data = token_util.EvalDataset(
            text_data,
            context_transform=context_transform,
            candidate_transform=None,
            mode=embed_type,
        )
    else:
        candidate_transform = token_util.SelectionSequentialTransform(
            tokenizer=tokenizer, max_len=train_config["max_candidate_length"]
        )
        context_data = token_util.EvalDataset(
            text_data,
            context_transform=None,
            candidate_transform=candidate_transform,
            mode=embed_type,
        )
    dataloader = DataLoader(
        context_data,
        batch_size=train_config["eval_batch_size"],
        collate_fn=context_data.eval_str,
        shuffle=False,
        num_workers=0,
    )

    for step, batch in enumerate(dataloader, start=1):
        batch = tuple(t.to(train_config["device"]) for t in batch)
        token_ids_batch, input_masks_batch = batch

        with torch.no_grad():
            if embed_type == "context":
                embeddings = model(
                    context_input_ids=token_ids_batch,
                    context_input_masks=input_masks_batch,
                    get_embedding=embed_type,
                    mode="get_embed",
                )
            else:
                embeddings = model(
                    candidate_input_ids=token_ids_batch,
                    candidate_input_masks=input_masks_batch,
                    get_embedding=embed_type,
                    mode="get_embed",
                )

    return embeddings.squeeze(0).detach().tolist()
